import {
  COMMENT_NODE,
  DOCUMENT_NODE,
  DOCUMENT_TYPE_NODE,
  ELEMENT_NODE,
  TEXT_NODE,
} from "../constants";
import { normalizeOutlookDocument } from "./strategies/outlook";
import { EMAIL_VENDORS } from "../constants";
import { normalizeTextSpacings } from "../utils";
import { normalizeGmailDocument } from "./strategies/gmail";
import { PlainNormalizer } from "../index";
import {
  ATTRIBUTES_TO_KEEP,
  normalizeDocumentCommon,
  pruneElement,
} from "./strategies/common";

const documentNormalizationFunctions = {
  [EMAIL_VENDORS.GMAIL]: normalizeGmailDocument,
  [EMAIL_VENDORS.OUTLOOK]: normalizeOutlookDocument,
};

export const normalizeVendorHtml = (
  document: HTMLDocument,
  vendor: string
): string => {
  /**
   * Remove unnecessary nodes
   */
  pruneHtmlNode(document.body);

  /**
   * Apply document normalisations
   */
  normalizeDocumentCommon(document.body);
  const normalizeDocument = documentNormalizationFunctions[vendor];
  if (normalizeDocument) {
    normalizeDocument(document);
  }

  /**
   * Final printout
   */
  return printHtmlChildren(document.body, 0);
};

export const extractPseudoPlainPart = (document: HTMLDocument): string => {
  let normalizedTextContent = PlainNormalizer.normalizePlain(
    document.body.textContent
  );

  const anchors = document.getElementsByTagName("a");
  const images = document.getElementsByTagName("img");
  let urls = [];

  Array.from(anchors).forEach((a) => {
    urls.push(a.getAttribute("href"));
  });
  Array.from(images).forEach((img) => {
    urls.push(img.getAttribute("src"));
  });

  urls = urls.filter((attr) => !!attr).sort();

  normalizedTextContent += urls.join(",");

  return normalizedTextContent;
};

export const pruneHtmlNode = (node: Node): boolean => {
  let toBeRemoved = false;

  switch (node.nodeType) {
    case COMMENT_NODE:
    case DOCUMENT_TYPE_NODE:
      toBeRemoved = true;
      break;
    case ELEMENT_NODE:
      toBeRemoved = pruneElement(node as HTMLElement);
      break;
  }

  if (toBeRemoved) {
    return true;
  }

  const childrenToRemove = [];
  let child = node.firstChild;
  while (child) {
    pruneHtmlNode(child) && childrenToRemove.push(child);
    child = child.nextSibling;
  }

  childrenToRemove.forEach((child) => node.removeChild(child));

  return false;
};

export const printHtmlChildren = (node: Node, depth: number): string => {
  let child = node.firstChild;
  if (!child) {
    return "";
  }

  if (child == node.lastChild && child.nodeType == TEXT_NODE) {
    return printHtmlNode(child, depth);
  } else {
    let result = "";
    while (child) {
      result = result.concat(printHtmlNode(child, depth));
      child = child.nextSibling;
    }

    return result;
  }
};

export const printHtmlNode = (node: Node, depth: number): string => {
  let result = "";

  switch (node.nodeType) {
    case TEXT_NODE: {
      const text = normalizeTextSpacings(node.textContent).trim();
      if (text.length) {
        result += "<TEXT>";
        result += text;
        result += "</TEXT>";
        result += "\n";
      }
      break;
    }
    case DOCUMENT_NODE:
      result += printHtmlChildren(node, depth);
      break;
    case ELEMENT_NODE:
      result += "<" + node.nodeName;
      Array.from((node as HTMLElement).attributes)
        .filter((a) => ATTRIBUTES_TO_KEEP[a.name])
        .sort((a, b) => a.name.localeCompare(b.name))
        .forEach((attribute) => {
          result += ` ${attribute.name}`;
          if (attribute.value) {
            result += `="${attribute.value}"`;
          }
        });

      if (node.firstChild) {
        const printout = printHtmlChildren(node, depth + 1);
        if (printout.trim().length === 0) {
          result += "/>";
        } else {
          result += ">";
          result += "\n";
          result += printout;
          result += "</" + node.nodeName + ">";
        }
      } else {
        result += "/>";
      }
      result += "\n";
      break;
  }

  return result;
};
